@@ -18,28 +18,39 @@ Copyright (C) 2010 Hiroki Ohtani(liris)
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 """
+from __future__ import print_function
 
 
+import six
 import socket
+
 try:
     import ssl
     from ssl import SSLError
-    from ssl import match_hostname
+    if hasattr(ssl, "match_hostname"):
+        from ssl import match_hostname
+    else:
+        from backports.ssl_match_hostname import match_hostname
+
     HAVE_SSL = True
 except ImportError:
-    HAVE_SSL = False
-
-    # dummy class for SSLError
+    # dummy class of SSLError for ssl none-support environment.
     class SSLError(Exception):
         pass
 
-from urllib.parse import urlparse
+    HAVE_SSL = False
+
+from six.moves.urllib.parse import urlparse
+if six.PY3:
+    from base64 import encodebytes as base64encode
+else:
+    from base64 import encodestring as base64encode
+
 import os
 import array
 import struct
 import uuid
 import hashlib
-import base64
 import threading
 import time
 import logging
@@ -153,7 +164,7 @@ def _parse_url(url):
 
     scheme, url = url.split(":", 1)
 
-    parsed = urlparse(url, scheme="http")
+    parsed = urlparse(url, scheme="ws")
     if parsed.hostname:
         hostname = parsed.hostname
     else:
@@ -202,8 +213,11 @@ def create_connection(url, timeout=None, **options):
     timeout: socket timeout time. This value is integer.
              if you set None for this value, it means "use default_timeout value"
 
-    options: current support option is only "header".
-             if you set header as dict value, the custom HTTP headers are added.
+
+    options: "header" -> custom http header list.
+             "cookie" -> cookie value.
+             "http_proxy_host" - http proxy host name.
+             "http_proxy_port" - http proxy port. If not set, set to 80.
     """
     sockopt = options.get("sockopt", [])
     sslopt = options.get("sslopt", {})
@@ -223,7 +237,7 @@ _MAX_CHAR_BYTE = (1<<8) -1
 
 def _create_sec_websocket_key():
     uid = uuid.uuid4()
-    return base64.encodebytes(uid.bytes).strip().decode("utf-8")
+    return base64encode(uid.bytes).decode('utf-8').strip()
 
 
 _HEADERS_TO_CHECK = {
@@ -299,7 +313,7 @@ class ABNF(object):
 
         fin: fin flag. if set to 0, create continue fragmentation.
         """
-        if opcode == ABNF.OPCODE_TEXT and isinstance(data, str):
+        if opcode == ABNF.OPCODE_TEXT and isinstance(data, six.text_type):
             data = data.encode("utf-8")
         # mask must be set if send data from client
         return ABNF(fin, 0, 0, 0, opcode, 1, data)
@@ -316,17 +330,19 @@ class ABNF(object):
         if length >= ABNF.LENGTH_63:
             raise ValueError("data is too long")
 
-        frame_header = []
-        frame_header = bytes((self.fin << 7
+        frame_header = chr(self.fin << 7
                            | self.rsv1 << 6 | self.rsv2 << 5 | self.rsv3 << 4
-                           | self.opcode,))
+                           | self.opcode)
         if length < ABNF.LENGTH_7:
-            frame_header += bytes((self.mask << 7 | length,))
+            frame_header += chr(self.mask << 7 | length)
+            frame_header = six.b(frame_header)
         elif length < ABNF.LENGTH_16:
-            frame_header += bytes((self.mask << 7 | 0x7e,))
+            frame_header += chr(self.mask << 7 | 0x7e)
+            frame_header = six.b(frame_header)
             frame_header += struct.pack("!H", length)
         else:
-            frame_header += bytes((self.mask << 7 | 0x7f,))
+            frame_header += chr(self.mask << 7 | 0x7f)
+            frame_header = six.b(frame_header)
             frame_header += struct.pack("!Q", length)
 
         if not self.mask:
@@ -337,6 +353,10 @@ class ABNF(object):
 
     def _get_masked(self, mask_key):
         s = ABNF.mask(mask_key, self.data)
+
+        if isinstance(mask_key, six.text_type):
+            mask_key = mask_key.encode('utf-8')
+
         return mask_key + s
 
     @staticmethod
@@ -348,12 +368,22 @@ class ABNF(object):
 
         data: data to mask/unmask.
         """
+
+        if isinstance(mask_key, six.text_type):
+            mask_key = six.b(mask_key)
+
+        if isinstance(data, six.text_type):
+            data = six.b(data)
+
         _m = array.array("B", mask_key)
         _d = array.array("B", data)
         for i in range(len(_d)):
             _d[i] ^= _m[i % 4]
 
-        return _d.tobytes()
+        if six.PY3:
+            return _d.tobytes()
+        else:
+            return _d.tostring()
 
 
 class WebSocket(object):
@@ -392,12 +422,9 @@ class WebSocket(object):
         if sslopt is None:
             sslopt = {}
         self.connected = False
-        self.sock = socket.socket()
-        self.raw_sock = None
-        for opts in DEFAULT_SOCKET_OPTION:
-            self.sock.setsockopt(*opts)
-        for opts in sockopt:
-            self.sock.setsockopt(*opts)
+        self.sock = None
+        self._timeout = None
+        self.sockopt = sockopt
         self.sslopt = sslopt
         self.get_mask_key = get_mask_key
         self.fire_cont_frame = fire_cont_frame
@@ -429,7 +456,7 @@ class WebSocket(object):
         """
         Get the websocket timeout(second).
         """
-        return self.sock.gettimeout()
+        return self._timeout
 
     def settimeout(self, timeout):
         """
@@ -437,7 +464,7 @@ class WebSocket(object):
 
         timeout: timeout time(second).
         """
-        self.sock.settimeout(timeout)
+        self._timeout = timeout
 
     timeout = property(gettimeout, settimeout)
 
@@ -445,29 +472,53 @@ class WebSocket(object):
         """
         Connect to url. url is websocket url scheme. ie. ws://host:port/resource
         You can customize using 'options'.
-        If you set "header" dict object, you can set your own custom header.
+        If you set "header" list object, you can set your own custom header.
 
         >>> ws = WebSocket()
         >>> ws.connect("ws://echo.websocket.org/",
-                ...     header={"User-Agent: MyProgram",
-                ...             "x-custom: header"})
+                ...     header=["User-Agent: MyProgram",
+                ...             "x-custom: header"])
 
         timeout: socket timeout time. This value is integer.
                  if you set None for this value,
                  it means "use default_timeout value"
 
-        options: current support option is only "header".
-                 if you set header as dict value,
-                 the custom HTTP headers are added.
+        options: "header" -> custom http header list.
+                 "cookie" -> cookie value.
+                 "http_proxy_host" - http proxy host name.
+                 "http_proxy_port" - http proxy port. If not set, set to 80.
 
         """
+
         hostname, port, resource, is_secure = _parse_url(url)
+        proxy_host, proxy_port = options.get("http_proxy_host", None), options.get("http_proxy_port", 0)
+        if not proxy_host:
+            addrinfo_list = socket.getaddrinfo(hostname, port, 0, 0, socket.SOL_TCP)
+        else:
+            proxy_port = proxy_port and proxy_port or 80
+            addrinfo_list = socket.getaddrinfo(proxy_host, proxy_port, 0, 0, socket.SOL_TCP)
+
+        if not addrinfo_list:
+            raise WebSocketException("Host not found.: " + hostname + ":" + str(port))
+
+        family = addrinfo_list[0][0]
+        self.sock = socket.socket(family)
+        self.sock.settimeout(self.timeout)
+        for opts in DEFAULT_SOCKET_OPTION:
+            self.sock.setsockopt(*opts)
+        for opts in self.sockopt:
+            self.sock.setsockopt(*opts)
         # TODO: we need to support proxy
-        self.sock.connect((hostname, port))
+        address = addrinfo_list[0][4]
+        self.sock.connect(address)
+
+        if proxy_host:
+            self._tunnel(hostname, port)
+
         if is_secure:
             if HAVE_SSL:
                 sslopt = dict(cert_reqs=ssl.CERT_REQUIRED,
-                    ca_certs=os.path.join(os.path.dirname(__file__), "cacert.pem"))
+                              ca_certs=os.path.join(os.path.dirname(__file__), "cacert.pem"))
                 sslopt.update(self.sslopt)
                 self.sock = ssl.wrap_socket(self.sock, **sslopt)
                 match_hostname(self.sock.getpeercert(), hostname)
@@ -476,6 +527,21 @@ class WebSocket(object):
 
         self._handshake(hostname, port, resource, **options)
 
+    def _tunnel(self, host, port):
+        logger.debug("Connecting proxy...")
+        connect_header = "CONNECT %s:%d HTTP/1.0\r\n" % (host, port)
+        connect_header += "\r\n"
+        if traceEnabled:
+            logger.debug("--- request header ---")
+            logger.debug(connect_header)
+            logger.debug("-----------------------")
+
+        self._send(connect_header)
+
+        status, resp_headers = self._read_headers()
+        if status != 200:
+            raise WebSocketException("failed CONNECT via proxy")
+
     def _handshake(self, host, port, resource, **options):
         headers = []
         headers.append("GET %s HTTP/1.1" % resource)
@@ -495,14 +561,18 @@ class WebSocket(object):
         key = _create_sec_websocket_key()
         headers.append("Sec-WebSocket-Key: %s" % key)
         headers.append("Sec-WebSocket-Version: %s" % VERSION)
+        
         if "header" in options:
             headers.extend(options["header"])
+        cookie = options.get("cookie", None)
+        if cookie:
+          headers.append("Cookie: %s" % cookie)
 
         headers.append("")
         headers.append("")
 
         header_str = "\r\n".join(headers)
-        self._send(bytes(header_str, "utf-8"))
+        self._send(header_str)
         if traceEnabled:
             logger.debug("--- request header ---")
             logger.debug(header_str)
@@ -534,9 +604,11 @@ class WebSocket(object):
             return False
         result = result.lower()
 
-        value = key + "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
-        digest = hashlib.sha1(bytes(value, "utf-8")).digest()
-        hashed = base64.encodebytes(digest).strip().decode("utf-8").lower()
+        if isinstance(result, six.text_type):
+            result = result.encode('utf-8')
+
+        value = (key + "258EAFA5-E914-47DA-95CA-C5AB0DC85B11").encode('utf-8')
+        hashed = base64encode(hashlib.sha1(value).digest()).strip().lower()
         return hashed == result
 
     def _read_headers(self):
@@ -544,9 +616,11 @@ class WebSocket(object):
         headers = {}
         if traceEnabled:
             logger.debug("--- response header ---")
+
         while True:
             line = self._recv_line()
-            if line == "\r\n":
+            line = line.decode('utf-8')
+            if line == "\r\n" or line == "\n":
                 break
             line = line.strip()
             if traceEnabled:
@@ -571,14 +645,13 @@ class WebSocket(object):
         """
         Send the data as string.
 
-        payload: Payload must be utf-8 bytes or str,
+        payload: Payload must be utf-8 string or unicoce,
                   if the opcode is OPCODE_TEXT.
-                  Otherwise, it must be byte array
+                  Otherwise, it must be string(byte array)
 
         opcode: operation code to send. Please see OPCODE_XXX.
         """
-        if isinstance(payload, str):
-            payload = bytes(payload, "utf-8")
+
         frame = ABNF.create_frame(payload, opcode)
         return self.send_frame(frame)
 
@@ -608,6 +681,7 @@ class WebSocket(object):
             data = data[l:]
         return length
 
+
     def send_binary(self, payload):
         return self.send(payload, ABNF.OPCODE_BINARY)
 
@@ -634,7 +708,7 @@ class WebSocket(object):
         return value: string(byte array) value.
         """
         opcode, data = self.recv_data()
-        if opcode == ABNF.OPCODE_TEXT:
+        if six.PY3 and opcode == ABNF.OPCODE_TEXT:
             return data.decode("utf-8")
         return data
 
@@ -720,17 +794,26 @@ class WebSocket(object):
 
         return value: ABNF frame object.
         """
+
         # Header
         if self._frame_header is None:
             self._frame_header = self._recv_strict(2)
+
         b1 = self._frame_header[0]
+
+        if six.PY2:
+            b1 = ord(b1)
+
         fin = b1 >> 7 & 1
         rsv1 = b1 >> 6 & 1
         rsv2 = b1 >> 5 & 1
         rsv3 = b1 >> 4 & 1
         opcode = b1 & 0xf
-
         b2 = self._frame_header[1]
+
+        if six.PY2:
+            b2 = ord(b2)
+
         has_mask = b2 >> 7 & 1
 
         # Frame length
@@ -761,18 +844,17 @@ class WebSocket(object):
 
         return ABNF(fin, rsv1, rsv2, rsv3, opcode, has_mask, payload)
 
-    def send_close(self, status=STATUS_NORMAL, reason=b""):
+
+    def send_close(self, status=STATUS_NORMAL, reason=""):
         """
         send close data to the server.
 
         status: status code to send. see STATUS_XXX.
 
-        reason: the reason to close. This must be string or bytes
+        reason: the reason to close. This must be string.
         """
         if status < 0 or status >= ABNF.LENGTH_16:
             raise ValueError("code is invalid range")
-        if isinstance(reason, str):
-            reason = bytes(reason, "utf-8")
         self.send(struct.pack('!H', status) + reason, ABNF.OPCODE_CLOSE)
 
     def close(self, status=STATUS_NORMAL, reason=""):
@@ -810,16 +892,34 @@ class WebSocket(object):
         self.sock.close()
 
     def _send(self, data):
+
+        if isinstance(data, six.text_type):
+            data = data.encode('utf-8')
+
         try:
             return self.sock.send(data)
         except socket.timeout as e:
-            raise WebSocketTimeoutException(*e.args)
+            message = getattr(e, 'strerror', getattr(e, 'message', ''))
+            raise WebSocketTimeoutException(message)
+        except Exception as e:
+            message = getattr(e, 'strerror', getattr(e, 'message', ''))
+            if "timed out" in message:
+                raise WebSocketTimeoutException(message)
+            else:
+                raise
 
     def _recv(self, bufsize):
         try:
             bytes = self.sock.recv(bufsize)
         except socket.timeout as e:
-            raise WebSocketTimeoutException(*e.args)
+            message = getattr(e, 'strerror', getattr(e, 'message', ''))
+            raise WebSocketTimeoutException(message)
+        except SSLError as e:
+            message = getattr(e, 'strerror', getattr(e, 'message', ''))
+            if message == "The read operation timed out":
+                raise WebSocketTimeoutException(message)
+            else:
+                raise
 
         if not bytes:
             raise WebSocketConnectionClosedException()
@@ -832,7 +932,8 @@ class WebSocket(object):
             self._recv_buffer.append(bytes)
             shortage -= len(bytes)
 
-        unified = b"".join(self._recv_buffer)
+        unified = six.b("").join(self._recv_buffer)
+
         if shortage == 0:
             self._recv_buffer = []
             return unified
@@ -840,14 +941,15 @@ class WebSocket(object):
             self._recv_buffer = [unified[bufsize:]]
             return unified[:bufsize]
 
+
     def _recv_line(self):
         line = []
         while True:
             c = self._recv(1)
             line.append(c)
-            if c == b"\n":
+            if c == six.b("\n"):
                 break
-        return b"".join(line).decode("utf-8")
+        return six.b("").join(line)
 
 
 class WebSocketApp(object):
@@ -859,7 +961,7 @@ class WebSocketApp(object):
                  on_open=None, on_message=None, on_error=None,
                  on_close=None, on_ping=None, on_pong=None,
                  on_cont_message=None,
-                 keep_running=True, get_mask_key=None):
+                 keep_running=True, get_mask_key=None, cookie=None):
         """
         url: websocket url.
         header: custom header for websocket handshake.
@@ -887,6 +989,7 @@ class WebSocketApp(object):
         """
         self.url = url
         self.header = header
+        self.cookie = cookie
         self.on_open = on_open
         self.on_message = on_message
         self.on_error = on_error
@@ -905,6 +1008,7 @@ class WebSocketApp(object):
         data: message to send. If you set opcode to OPCODE_TEXT, data must be utf-8 string or unicode.
         opcode: operation code of data. default is OPCODE_TEXT.
         """
+
         if self.sock.send(data, opcode) == 0:
             raise WebSocketConnectionClosedException()
 
@@ -920,7 +1024,8 @@ class WebSocketApp(object):
             self.last_ping_tm = time.time()
             self.sock.ping()
 
-    def run_forever(self, sockopt=None, sslopt=None, ping_interval=0, ping_timeout=None):
+    def run_forever(self, sockopt=None, sslopt=None, ping_interval=0, ping_timeout=None,
+        http_proxy_host=None, http_proxy_port=None):
         """
         run event loop for WebSocket framework.
         This loop is infinite loop and is alive during websocket is available.
@@ -930,7 +1035,10 @@ class WebSocketApp(object):
         ping_interval: automatically send "ping" command every specified period(second)
             if set to 0, not send automatically.
         ping_timeout: timeout(second) if the pong message is not recieved.
+        http_proxy_host: http proxy host name.
+        http_proxy_port: http proxy port. If not set, set to 80.
         """
+
         if not ping_timeout or ping_timeout<=0:
             ping_timeout = None
         if sockopt is None:
@@ -944,7 +1052,9 @@ class WebSocketApp(object):
         try:
             self.sock = WebSocket(self.get_mask_key, sockopt=sockopt, sslopt=sslopt,
                 fire_cont_frame=self.on_cont_message and True or False)
-            self.sock.connect(self.url, header=self.header)
+            self.sock.settimeout(default_timeout)
+            self.sock.connect(self.url, header=self.header, cookie=self.cookie,
+                http_proxy_host=http_proxy_host, http_proxy_port=http_proxy_port)
             self._callback(self.on_open)
 
             if ping_interval:
@@ -960,7 +1070,7 @@ class WebSocketApp(object):
                 if ping_timeout and self.last_ping_tm and time.time() - self.last_ping_tm > ping_timeout:
                     self.last_ping_tm = 0
                     raise WebSocketTimeoutException()
-                    
+
                 if r:
                     op_code, frame = self.sock.recv_data_frame(True)
                     if op_code == ABNF.OPCODE_CLOSE:
@@ -973,7 +1083,7 @@ class WebSocketApp(object):
                         self._callback(self.on_cont_message, frame.data, frame.fin)
                     else:
                         data = frame.data
-                        if frame.opcode == ABNF.OPCODE_TEXT:
+                        if six.PY3 and frame.opcode == ABNF.OPCODE_TEXT:
                             data = data.decode("utf-8")
                         self._callback(self.on_message, data)
         except Exception as e:
